/*
 * ValueOperationAbstraction.hpp
 *
 *  Created on: 11. 7. 2017
 *	  Author: Jan Travnicek
 */

#ifndef _VALUE_OPERATION_ABSTRACTION_HPP_
#define _VALUE_OPERATION_ABSTRACTION_HPP_

#include <alib/tuple>
#include <alib/memory>
#include <alib/typeindex>

#include <abstraction/OperationAbstraction.hpp>

namespace abstraction {

template < class ReturnType >
class ValueInterface : public virtual OperationAbstraction {
public:
	virtual ReturnType && getValue ( ) const = 0;

	virtual bool isConst ( ) const = 0;
	virtual bool isRvalueRef ( ) const = 0;
	virtual bool isLvalueRef ( ) const = 0;
	virtual bool isAutoMove ( ) const {
		return false;
	}

	ext::type_index getReturnTypeIndex ( ) const override {
		return ext::type_index ( typeid ( ReturnType ) );
	}

};

} /* namespace abstraction */

#include <common/ParamQualifiers.hpp>
#include <common/AbstractionHelpers.hpp>

namespace abstraction {

template < class ReturnType >
class ValueOperationAbstraction : virtual public ValueInterface < ReturnType > {
	mutable std::optional < ReturnType > m_data;

protected:
	bool isConst ( ) const override {
		return false;
	}

	bool isRvalueRef ( ) const override {
		return false;
	}

	bool isLvalueRef ( ) const override {
		return false;
	}

	void setData ( ReturnType data ) {
		m_data = std::move ( data );
	}

	ReturnType && getValue ( ) const override {
		return std::move ( m_data.value ( ) );
	}

public:
	template < typename ... ParamTypes, typename Callable >
	inline void run_helper ( Callable callback, const ext::array < std::pair < std::shared_ptr < OperationAbstraction >, bool >, sizeof ... ( ParamTypes ) > & inputs ) {
		if ( ! evaluated ( ) )
			setData ( abstraction::apply < ParamTypes ... > ( callback, inputs ) );
	}

	ext::set < abstraction::ParamQualifiers::ParamQualifier > getReturnTypeQualifiers ( ) const override {
		return abstraction::ParamQualifiers::paramQualifiers < ReturnType > ( );
	}

	bool evaluated ( ) const override {
		return ( bool ) m_data;
	}
};

template < class ReturnType >
class ValueOperationAbstraction < const ReturnType > : virtual public ValueInterface < ReturnType > {
	mutable std::optional < ReturnType > m_data;

protected:
	bool isConst ( ) const override {
		return true;
	}

	bool isRvalueRef ( ) const override {
		return false;
	}

	bool isLvalueRef ( ) const override {
		return false;
	}

	void setData ( ReturnType data ) {
		m_data = std::move ( data );
	}

	ReturnType && getValue ( ) const override {
		return std::move ( m_data.value ( ) );
	}

public:
	template < typename ... ParamTypes, typename Callable >
	inline void run_helper ( Callable callback, const ext::array < std::pair < std::shared_ptr < OperationAbstraction >, bool >, sizeof ... ( ParamTypes ) > & inputs ) {
		if ( ! evaluated ( ) )
			setData ( abstraction::apply < ParamTypes ... > ( callback, inputs ) );
	}

	ext::set < abstraction::ParamQualifiers::ParamQualifier > getReturnTypeQualifiers ( ) const override {
		return abstraction::ParamQualifiers::paramQualifiers < const ReturnType > ( );
	}

	bool evaluated ( ) const override {
		return ( bool ) m_data;
	}
};

template < class ReturnType >
class ValueOperationAbstraction < ReturnType & > : virtual public ValueInterface < ReturnType > {
	mutable std::optional < std::reference_wrapper < ReturnType > > m_data;

protected:
	bool isConst ( ) const override {
		return false;
	}

	bool isRvalueRef ( ) const override {
		return true;
	}

	bool isLvalueRef ( ) const override {
		return false;
	}

	void setData ( ReturnType & data ) {
		m_data = std::reference_wrapper < ReturnType > ( data );
	}

	ReturnType && getValue ( ) const override {
		return std::move ( m_data->get ( ) );
	}

public:
	template < typename ... ParamTypes, typename Callable >
	inline void run_helper ( Callable callback, const ext::array < std::pair < std::shared_ptr < OperationAbstraction >, bool >, sizeof ... ( ParamTypes ) > & inputs ) {
		if ( ! evaluated ( ) )
			setData ( abstraction::apply < ParamTypes ... > ( callback, inputs ) );
	}

	ext::set < abstraction::ParamQualifiers::ParamQualifier > getReturnTypeQualifiers ( ) const override {
		return abstraction::ParamQualifiers::paramQualifiers < ReturnType & > ( );
	}

	bool evaluated ( ) const override {
		return ( bool ) m_data;
	}
};

template < class ReturnType >
class ValueOperationAbstraction < const ReturnType & > : virtual public ValueInterface < ReturnType > {
	mutable std::optional < std::reference_wrapper < const ReturnType > > m_data;

protected:
	bool isConst ( ) const override {
		return true;
	}

	bool isRvalueRef ( ) const override {
		return true;
	}

	bool isLvalueRef ( ) const override {
		return false;
	}

	void setData ( const ReturnType & data ) {
		m_data = std::reference_wrapper < const ReturnType > ( data );
	}

	ReturnType && getValue ( ) const override {
		return std::move ( const_cast < ReturnType & > ( m_data->get ( ) ) );
	}

public:
	template < typename ... ParamTypes, typename Callable >
	inline void run_helper ( Callable callback, const ext::array < std::pair < std::shared_ptr < OperationAbstraction >, bool >, sizeof ... ( ParamTypes ) > & inputs ) {
		if ( ! evaluated ( ) )
			setData ( abstraction::apply < ParamTypes ... > ( callback, inputs ) );
	}

	ext::set < abstraction::ParamQualifiers::ParamQualifier > getReturnTypeQualifiers ( ) const override {
		return abstraction::ParamQualifiers::paramQualifiers < const ReturnType & > ( );
	}

	bool evaluated ( ) const override {
		return ( bool ) m_data;
	}
};

template < class ReturnType >
class ValueOperationAbstraction < ReturnType && > : virtual public ValueInterface < ReturnType > {
	mutable std::optional < std::reference_wrapper < ReturnType > > m_data;

protected:
	bool isConst ( ) const override {
		return false;
	}

	bool isRvalueRef ( ) const override {
		return false;
	}

	bool isLvalueRef ( ) const override {
		return true;
	}

	void setData ( ReturnType && data ) {
		m_data = std::reference_wrapper < ReturnType > ( data );
	}

	ReturnType && getValue ( ) const override {
		return std::move ( m_data->get ( ) );
	}

public:
	template < typename ... ParamTypes, typename Callable >
	inline void run_helper ( Callable callback, const ext::array < std::pair < std::shared_ptr < OperationAbstraction >, bool >, sizeof ... ( ParamTypes ) > & inputs ) {
		if ( ! evaluated ( ) )
			setData ( abstraction::apply < ParamTypes ... > ( callback, inputs ) );
	}

	ext::set < abstraction::ParamQualifiers::ParamQualifier > getReturnTypeQualifiers ( ) const override {
		return abstraction::ParamQualifiers::paramQualifiers < ReturnType && > ( );
	}

	bool evaluated ( ) const override {
		return ( bool ) m_data;
	}
};

template < class ReturnType >
class ValueOperationAbstraction < const ReturnType && > : virtual public ValueInterface < ReturnType > {
	mutable std::optional < std::reference_wrapper < const ReturnType > > m_data;

protected:
	bool isConst ( ) const override {
		return true;
	}

	bool isRvalueRef ( ) const override {
		return false;
	}

	bool isLvalueRef ( ) const override {
		return true;
	}

	void setData ( const ReturnType && data ) {
		m_data = std::reference_wrapper < const ReturnType > ( data );
	}

	ReturnType && getValue ( ) const override {
		return std::move ( const_cast < ReturnType & > ( m_data->get ( ) ) );
	}

public:
	template < typename ... ParamTypes, typename Callable >
	inline void run_helper ( Callable callback, const ext::array < std::pair < std::shared_ptr < OperationAbstraction >, bool >, sizeof ... ( ParamTypes ) > & inputs ) {
		if ( ! evaluated ( ) )
			setData ( abstraction::apply < ParamTypes ... > ( callback, inputs ) );
	}

	ext::set < abstraction::ParamQualifiers::ParamQualifier > getReturnTypeQualifiers ( ) const override {
		return abstraction::ParamQualifiers::paramQualifiers < const ReturnType && > ( );
	}

	bool evaluated ( ) const override {
		return ( bool ) m_data;
	}
};

template < >
class ValueOperationAbstraction < void > : virtual public OperationAbstraction {
public:
	template < typename ... ParamTypes, typename Callable >
	inline void run_helper ( Callable callback, const ext::array < std::pair < std::shared_ptr < OperationAbstraction >, bool >, sizeof ... ( ParamTypes ) > & inputs ) {
		abstraction::apply < ParamTypes ... > ( callback, inputs );
	}

	ext::type_index getReturnTypeIndex ( ) const override {
		return ext::type_index ( typeid ( void ) );
	}

	ext::set < abstraction::ParamQualifiers::ParamQualifier > getReturnTypeQualifiers ( ) const override {
		return abstraction::ParamQualifiers::paramQualifiers < void > ( );
	}

	bool evaluated ( ) const override {
		return false;
	}
};

} /* namespace abstraction */

#endif /* _VALUE_OPERATION_ABSTRACTION_HPP_ */
