#ifndef _ABSTRACTION_HELPERS_HPP_
#define _ABSTRACTION_HELPERS_HPP_

#include <alib/tuple>
#include <alib/array>
#include <alib/typeinfo>

namespace abstraction {

template < class ParamType >
ParamType retrieveValue ( const std::shared_ptr < OperationAbstraction > & param, bool move ) {
	(void) move;
	using Type = std::decay_t < ParamType >;

	std::shared_ptr < ValueInterface < std::decay_t < Type > > > interface = std::dynamic_pointer_cast < ValueInterface < std::decay_t < Type > > > ( param->getProxyAbstraction ( ) );
	if ( ! interface )
		throw std::invalid_argument ( "Abstraction does not provide value of type " + ext::to_string < ParamType > ( ) + " but " + param->getReturnType ( ) + "." );

	if constexpr ( std::is_rvalue_reference_v < ParamType > ) {
		if ( move || interface->isAutoMove ( ) )
			return std::move ( interface->getValue ( ) );
		else
			throw std::domain_error ( "Cannot bind without move" );
	} else if constexpr ( std::is_lvalue_reference_v < ParamType > ) {
		Type && res = interface->getValue ( );
		return res;
	} else if constexpr ( std::is_copy_constructible_v < Type > && std::is_move_constructible_v < Type > ) {
		if ( ! interface->isConst ( ) && ( move || interface->isAutoMove ( ) ) )
			return std::move ( interface->getValue ( ) );
		else {
			Type && res = interface->getValue ( );
			return res;
		}
	} else if constexpr ( std::is_copy_constructible_v < Type > && ! std::is_move_constructible_v < Type > ) {
		if ( ! interface->isConst ( ) && move )
			throw std::domain_error ( "Value not move constructible" );
		else {
			Type && res = interface->getValue ( );
			return res;
		}
	} else if constexpr ( ! std::is_copy_constructible_v < Type > && std::is_move_constructible_v < Type > ) {
		if ( ! interface->isConst ( ) && move )
			return std::move ( interface->getValue ( ) );
		else
			throw std::domain_error ( "Value not copy constructible" );
	} else { // ! std::is_copy_constructible_v < Type > && ! std::is_move_constructible_v < Type >
		if ( ! interface->isConst ( ) && move )
			throw std::domain_error ( "Value not move constructible" );
		else
			throw std::domain_error ( "Value not copy constructible" );
	}
}

namespace detail {

template < class ... ParamTypes, class F, class Tuple, std::size_t... I >
constexpr decltype ( auto ) apply_impl ( F && f, Tuple && t, std::index_sequence < I ... > ) {
	return std::invoke ( std::forward < F > ( f ), abstraction::retrieveValue < ParamTypes > ( std::get < I > ( std::forward < Tuple > ( t ) ).first, std::get < I > ( std::forward < Tuple > ( t ) ).second ) ... );
}

}  // namespace detail

template < class ... ParamTypes, class F, class Tuple >
constexpr decltype ( auto ) apply ( F && f, Tuple && t ) {
	return detail::apply_impl < ParamTypes ... > ( std::forward < F > ( f ), std::forward < Tuple > ( t ), std::make_index_sequence < sizeof ... ( ParamTypes ) > { } );
}

template < class ... Params >
bool checkInput ( const std::shared_ptr < OperationAbstraction > & operation, unsigned index ) {
	bool res = false;

	auto lambda = [ & ] ( auto I ) {
		 res = ( bool ) std::dynamic_pointer_cast < std::decay_t < std::tuple_element_t < decltype ( I )::value, std::tuple < Params ... > > > > ( operation->getProxyAbstraction ( ) );
	};

	ext::constexpr_switch < sizeof ... ( Params ) > ( index, lambda );

	return res;
}

template < class ... Params >
static ext::type_index paramType ( unsigned index ) {
	ext::type_index res ( typeid ( void ) );

	auto lambda = [ & ] ( auto I ) {
		res = ext::type_index ( typeid ( std::tuple_element_t < decltype ( I )::value, std::tuple < Params ... > > ) );
	};

	ext::constexpr_switch < sizeof ... ( Params ) > ( index, lambda );

	return res;
}

template < class ... Params >
ext::set < abstraction::ParamQualifiers::ParamQualifier > paramTypeQualifiers ( unsigned index ) {
	ext::set < abstraction::ParamQualifiers::ParamQualifier > res;

	auto lambda = [ & ] ( auto I ) {
		res = abstraction::ParamQualifiers::paramQualifiers < std::tuple_element_t < decltype ( I )::value, std::tuple < Params ... > > > ( );
	};

	ext::constexpr_switch < sizeof ... ( Params ) > ( index, lambda );

	return res;
}

} /* namespace abstraction */

#endif // _ABSTRACTION_HELPERS_HPP_
